import sqlite3
from typing import Any, Dict, Generic, Iterable, Iterator, List, Tuple, Type, TypeVar, Union, overload

from ._core import BaseConnection, _IntegerParent, _FixedPointParent, IntegerField
from .exceptions import ConnectionError, DoesNotExistsError, ProgrammingError
from . import _translator

connection: BaseConnection = None
"""
database connection object.
"""

_created_models: List[object] = []
"""
created models.
"""

@overload
def connect(database: str, timeout: float = ..., detect_types: int = ..., isolation_level: Union[str, None] = None, check_same_thread: bool = ..., factory: Union[Type[sqlite3.Connection], None] = ..., cached_statements: int = ..., uri: bool = ...) -> BaseConnection:
    """
    create database connection object.

    https://docs.python.org/3/library/sqlite3.html#sqlite3.connect
    """
    pass

def connect(*args, **kwargs) -> BaseConnection:
    if not connection is None:
        raise ConnectionError(
            "You are already connected to database.", connection
        )

    globals()["connection"] = BaseConnection(*args, **kwargs)
    return connection

_ML = TypeVar("_ML")

class Model:

    objects: 'QuerySet[_ML]'
    _kwargs: Dict[str, Any]
    _primary_key: str = None
    _auto_increment: str = None

    def __init__(self, **kwargs) -> None:
        for k,v in kwargs.items():
            if k not in self.__slots__:
                raise ValueError(
                    "You want to set '%s' attribute to %r, but this attribute not created." % (k, v)
                )
            setattr(self, k, v)
        
        self._kwargs = kwargs
    
    def __init_subclass__(cls: Type[_ML]) -> None:
        cls.__model_fields__ = {i:(getattr(cls, i), getattr(cls, i).to_statement()) for i in dir(cls) if (not callable(getattr(cls,i)) and (not i.startswith('_')) and (i != "objects"))}

        cls.objects: QuerySet[_ML] = QuerySet(cls, [])
        
        for k, v in cls.__model_fields__.items():
            if v[0].primary_key == True:
                cls._primary_key = k
            
            if isinstance(v[0], (_IntegerParent)):
                if v[0].auto_increment == True:
                    cls._auto_increment = k
            
            setattr(cls, k, None)
        
        if cls._primary_key is None:
            field = IntegerField(primary_key=True, auto_increment=True, not_null=True)
            setattr(cls, 'id', None)
            cls.__model_fields__['id'] = (field, field.to_statement())
            cls._primary_key = 'id'
            cls._auto_increment = 'id'
        
        _created_models.append(cls)
        
        cls.__slots__ = tuple(i for i in cls.__model_fields__.keys())
    
    def save(self) -> None:
        """
        Insert new row to database.

        ### Example:
        ```python
        # ...
        user = Users(id=5, username="aWolver")
        user.save()
        ```
        """
        self = self.objects.manager.insert(self)
    
    def update(self, **kwargs) -> None:
        """
        update row in database.

        ### Args:
            - **kwargs (Any): new columns.
        
        ### Example:
        ```python
        # ...
        user = Users.objects.get(username="aWolver")
        user.update(username="Ali")
        ```
        """
        self.objects.manager.update(self, **kwargs)
        for k,v in kwargs.items():
            setattr(self, k, v)
            self._kwargs[k] = v

    def delete(self) -> None:
        """
        delete row from database.

        ### Example:
        ```python
        # ...
        user = Users.objects.get(username="aWolver")
        user.delete()
        ```
        """
        self.objects.manager.delete(self)
    
    def values_list(self, *args) -> Tuple[Any]:
        """
        ### Example:
        ```python
        # ...
        user = Users.objects.get(username="aWolver")
        print(user.values_list("id", "password"))
        # (1, 'PASSWORD')
        ```
        """
        return tuple(filter(lambda x: x[0] in args, self._kwargs.items()))

_T = TypeVar("_T")
_ML = TypeVar("_ML")

class QuerySet(Generic[_ML]):
    """
    A list of result rows.

    #### Can use for generic typing:
    ```python
    # ...
    all: eSQL.QuerySet[Users] = Users.objects.all()
    # ...
    ```
    """

    def __new__(cls: Type[_T], obj: Type[_ML], result: List[_ML] = None) -> _T:
        """
        ### Args:
            - obj (type(Model)):    your model.
            - result (list(Model)): result. (default: None)
        """
        return super().__new__(cls, obj, result or [])
    
    def __init__(self, obj: Type[_ML], result: List[_ML]) -> None:
        self.obj = obj
        self.manager = _Manager()
        self.result = result
    
    def first(self) -> Union[_ML, None]:
        """
        Return model if exists.
        """
        return None if self.empty() else self.result[0]
    
    def create(self, **kwargs: Any) -> _ML:
        """
        Insert to database.

        ### Args:
            - **kwargs (Any): columns.

        ### Raise:
        raise `ValueError` if not pass column parameters.  

        ### Example:
        ```
        # ...
        new_user = Users.objects.create(name="ali")
        # ...
        ```
        """
        _cls = self.obj(**kwargs)

        try:
            _cls = self.manager.insert(_cls)
        except ValueError:
            raise ValueError(
                'cannot create empty row, please pass parameters (%s model).' % (self.obj.__name__)
            ) from None
        return _cls

    def filter(self, _limit: int = None, **kwargs: Any) -> _T:
        """
        select rows from database.

        ### Args:
            - _limit (int):      selected limit.
            - **kwargs (Any):    statement condition.
        
        ### Return
        return QuerySet/List of model.
        """
        return self.manager.select(self.obj, _limit=_limit, _exclude=False, **kwargs)
    
    def exclude(self, _limit: int = None, **kwargs: Any) -> _T:
        """
        like filter(...) method but if you want to get database rows with negative condition, use this.
        """
        return self.manager.select(self.obj, _limit=_limit, _exclude=True, **kwargs)

    def get(self, **kwargs: Any) -> _ML:
        """
        like filter(...) but just return a row.

        ### Args:
            - **kwargs (Any): statement condition.
        
        ### Return
        return model.

        ### Raise
        raise `exceptions.DoesNotExistsError` if row doesn't exists.
        """
        res = self.manager.select(self.obj, **kwargs)
        if res.empty():
            raise DoesNotExistsError(
                'Not have a result for return.'
            )
        
        return res.first()

    def like(self, _limit: int = None, **kwargs: Any) -> _T:
        """
        like filter(...) but you can use sqlite `LIKE` statement.

        ### Args:
            - _limit (int):      selected limit.
            - **kwargs (Any):    statement condition.
        
        ### Return
        return QuerySet/List of model.

        ### Example:
        ```
        MODEL.objects.like(name="A%", country="%E")
        ```
        """
        return self.manager.like(self.obj, _limit=_limit, _exclude=False, **kwargs)

    def not_like(self, _limit: int = None, **kwargs: Any) -> _T:
        """
        like exclude(...) but you can use sqlite `NOT LIKE` statement.

        ### Args:
            - _limit (int):      selected limit.
            - **kwargs (Any):    statement condition.
        
        ### Return
        return QuerySet/List of model.

        ### Example:
        ```
        MODEL.objects.not_like(name="A%", country="%E")
        ```
        """
        return self.manager.like(self.obj, _limit=_limit, _exclude=True, **kwargs)

    def all(self, _limit: int = None) -> _T:
        """
        select all rows from database.

        ### Args:
            - _limit (int):      selected limit.
        
        ### Return
        return QuerySet/List of model.
        """
        return self.manager.select(self.obj, _limit=_limit)

    def count(self) -> int:
        """
        return result count.

        ### Note
        this method return length of getted result.

        if you want get length of rows in database, use len().
        """
        return len(self.result)

    def empty(self) -> bool:
        """
        return True if not result.
        """
        return len(self.result) == 0

    def raw(self, __sql: str, __parameters: Iterable = None) -> Any:
        """
        Raw quering.
        """
        return self.manager.raw(__sql, __parameters)

    def __len__(self) -> int:
        return self.manager.count(self.obj)[0]
    
    def __iter__(self) -> Iterator[Model]:
        return iter(self.result)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {list(str(i) for i in self.result)}>"

_T = TypeVar("_T")
_MLM = TypeVar("_MLM")

class _Manager:
    """
    (`private class`)

    Manage database connection object.
    """

    _instance = None
    def __new__(cls: Type[_T]) -> _T:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance
    
    def __init__(self) -> None:
        """
        (`private class`)
        database manager.

        ### Raise
        raise `exceptions.ConnectionError` if not connected to database.
        """
        try:
            self.cur = connection.cursor()
        except AttributeError:
            raise ConnectionError(
                'Manager not found connection. connect to database before create Models.'
            ) from None
    
    def create_table(self, *models: Type[Model], commit: bool = True, force: bool = False, verbose: bool = False) -> Union[None, str]:
        """
        create tables in database.

        ### Args:
            - *models (type(Model)): Your model classes.

            - commit (bool):  commit to database?
            - force (bool):   Dont use `IF NOT EXISTS` in sqlite statement.
            - verbose (bool): Return sqlite statement.
        """
        __sql = _translator.transaction()

        for mdl in models:
            __sql.add(_translator.create_table(mdl, force))
        
        __sql = str(__sql)

        if (commit): self.cur.executescript(str(__sql))

        if (verbose): return __sql

    def drop_table(self, *models: Type[Model], force: bool = False) -> None:
        """
        drop tables from database.

        ### Args:
            - *models (type(Model)): Your created model classes.
        """
        __sql = _translator.transaction()

        for mdl in models:
            __sql.add(_translator.drop_table(mdl, force))
        
        __sql = str(__sql)

        self.cur.executescript(__sql)

    def insert(self, model: _MLM) -> _MLM:
        """
        insert to database.

        ### Args:
            - model (Model): initialized model.
        """
        self.cur.execute(*_translator.insert_into.create(model.__class__, model._kwargs))
        
        mdl = model.__class__(**model._kwargs)

        if model._auto_increment:
            setattr(mdl, model._auto_increment, self.cur.lastrowid)

        return mdl

    def select(self, obj: Type[_MLM], _limit: int = None, _exclude: bool = False, **kwargs: Any) -> QuerySet[_MLM]:
        """
        select from database.

        ### Args:
            - obj (type(Model)): your model class.
            - _limit (int):      selected limit.
            - _exclude (bool):   if you want to get database rows with negative condition, set `True`.
            - **kwargs (Any):    statement condition.
        
        ### Return
        return `QuerySet` of selected rows.
        """
        operation, params = _translator.select(obj.__name__, kwargs, _limit=_limit, _exclude=_exclude).condition()

        self.cur.execute(operation, params)

        return QuerySet(obj, [obj(**i) for i in self.cur.fetchall()])

    def like(self, obj: Type[_MLM], _limit: int = None, _exclude: bool = False, **kwargs: Any) -> QuerySet[_MLM]:
        """
        select from database `use LIKE/NOT LIKE`.

        ### Args:
            - obj (type(Model)): your model class.
            - _limit (int):      selected limit.
            - _exclude (bool):   if you want to get database rows with negative condition, set `True`.
            - **kwargs (Any):    statement condition.
        
        ### Return
        return `QuerySet` of selected rows.
        """
        operation, params = _translator.select(obj.__name__, kwargs, _limit=_limit, _exclude=_exclude).like()

        self.cur.execute(operation, params)

        return QuerySet(obj, [obj(**i) for i in self.cur.fetchall()])

    def count(self, obj: Type[_MLM]) -> Tuple[int]:
        """
        select count of * from database.

        ### Args:
            - obj (type(Model)): your model class.
            - _limit (int):      selected limit.
            - _exclude (bool):   if you want to get database rows with negative condition, set `True`.
            - **kwargs (Any):    statement condition.
        
        ### Return
        return `QuerySet` of selected rows.
        """
        operation, params = _translator.select(obj.__name__).count()

        self.cur.execute(operation, params)

        return self.cur.fetchone()

    def delete(self, obj: Union[Model, QuerySet[Model]]) -> None:
        """
        delete row or rows from database.

        ### Args:
            - obj (Model | QuerySet): initialized model or QuerySet class.
        """
        if isinstance(obj, Model):
            self.cur.execute(*_translator.delete(obj.__class__.__name__, obj._kwargs))
            return
        
        elif isinstance(obj, QuerySet):
            for mdl in obj:
                self.cur.execute(*_translator.delete(mdl.__class__.__name__, mdl._kwargs))
            return
        
        else:
            raise TypeError(
                "_Manager.delete only accept 'eSQL.Model' or 'eSQL.QuerySet' type. got %s." % type(obj)
            )
    
    def update(self, obj: Union[Model, QuerySet[Model]], **kwargs: Any) -> None:
        """
        update row or rows from database.

        ### Args:
            - obj (Model | QuerySet): initialized model or QuerySet class.
            - **kwargs (Any):         `SET` parameters statement.
        """
        if isinstance(obj, Model):
            self.cur.execute(*_translator.update(obj.__class__.__name__, kwargs, obj._kwargs))
            return
        
        elif isinstance(obj, QuerySet):
            for mdl in obj:
                self.cur.execute(*_translator.update(mdl.__class__.__name__, kwargs, mdl._kwargs))
            return
        
        else:
            raise TypeError(
                "_Manager.update only accept 'eSQL.Model' or 'eSQL.QuerySet' type. got %s." % type(obj)
            )

    def raw(self, __sql: str, __parameters: Iterable = None) -> Any:
        """
        Raw quering.
        """
        self.cur.execute(__sql, __parameters)
        return self.cur.fetchall()

_T = TypeVar("_T")

class transaction:
    def __init__(self) -> None:
        if connection is None:
            raise ConnectionError(
                'Transaction not found connection. connect to database before using transaction.'
            ) from None
        connection.conn.isolation_level = ""
    
    def commit(self) -> None:
        try:connection.commit()
        finally:connection.conn.isolation_level = None
    
    def rollback(self, *args, **kwargs) -> None:
        try:connection.rollback(*args, **kwargs)
        finally:connection.conn.isolation_level = None

def sqlmigrate(*models: Type[Model], force: bool = False) -> str:
    """
    get SQL statement for create tables.

    ### Args:
        - *models (type(Model)): your models for create in database. if pass models, use all created models in database.
        - force (bool):          Use `IF NOT EXISTS` in sqlite statement.
    
    ### Example:
    ```
    import eSQL

    class Users(eSQL.Model):
       id = eSQL.IntegerField(primary_key=True)
    
    statement = eSQL.sqlmigrate()
    # or eSQL.sqlmigrate(Users)
    # if not pass parameters, use all created models.
    ```
    """
    models = models if models else _created_models
    return (_Manager().create_table(*models, force=force, commit=False, verbose=True))

def migrate(*models: Type[Model], force: bool = False) -> None:
    """
    Create tables in database.

    ### Args:
        - *models (type(Model)): your models for create in database. if pass models, create all created models in database.
        - force (bool):          Use `IF NOT EXISTS` in sqlite statement.
    
    ### Example:
    ```
    import eSQL

    class Users(eSQL.Model):
       id = eSQL.IntegerField(primary_key=True)
    
    eSQL.migrate()
    # or eSQL.migrate(Users)
    # if not pass parameters, use all created models.
    ```
    """
    models = models if models else _created_models
    _Manager().create_table(*models, force=force, commit=True, verbose=False)

def removemigrate(*models: Type[Model], force: bool = True):
    """
    Drop tables from database.

    ### Args:
        - *models (type(Model)): your models for create in database. if pass models, use all created models in database.
        - force (bool):          Use `IF EXISTS` in sqlite statement.

    ### Example:
    ```
    import eSQL

    class Users(eSQL.Model):
       id = eSQL.IntegerField(primary_key=True)
    
    eSQL.migrate()

    # CODE ...

    eSQL.removemigrate()
    # if not pass parameters, use all created models.
    ```
    """
    models = models if models else _created_models
    _Manager().drop_table(*models, force=force)
